@@ -24,6 +24,9 @@ use strict;
 use warnings;
 use re 'taint';
 
+require 5.008;     # need basic Unicode support for HTML::Parser::utf8_mode
+# require 5.008008;  # Bug 3787; [perl #37950]: Malformed UTF-8 character ...
+
 use HTML::Parser 3.43 ();
 use Mail::SpamAssassin::Logger;
 use Mail::SpamAssassin::Constants qw(:sa);
@@ -52,7 +55,7 @@ my %tricks = map {; $_ => 1 }
 
 # elements that change text style
 my %elements_text_style = map {; $_ => 1 }
-  qw( body font table tr th td big small basefont marquee span ),
+  qw( body font table tr th td big small basefont marquee span p div a ),
 ;
 
 # elements that insert whitespace
@@ -77,14 +80,17 @@ $ok_attributes{basefont}{$_} = 1 for qw( color face size );
 $ok_attributes{body}{$_} = 1 for qw( text bgcolor link alink vlink background );
 $ok_attributes{font}{$_} = 1 for qw( color face size );
 $ok_attributes{marquee}{$_} = 1 for qw( bgcolor background );
-$ok_attributes{table}{$_} = 1 for qw( bgcolor );
-$ok_attributes{td}{$_} = 1 for qw( bgcolor );
-$ok_attributes{th}{$_} = 1 for qw( bgcolor );
-$ok_attributes{tr}{$_} = 1 for qw( bgcolor );
+$ok_attributes{table}{$_} = 1 for qw( bgcolor style );
+$ok_attributes{td}{$_} = 1 for qw( bgcolor style );
+$ok_attributes{th}{$_} = 1 for qw( bgcolor style );
+$ok_attributes{tr}{$_} = 1 for qw( bgcolor style );
 $ok_attributes{span}{$_} = 1 for qw( style );
+$ok_attributes{p}{$_} = 1 for qw( style );
+$ok_attributes{div}{$_} = 1 for qw( style );
+$ok_attributes{a}{$_} = 1 for qw( style );
 
 sub new {
-  my ($class) = @_;
+  my ($class, $character_semantics_input, $character_semantics_output) = @_;
   my $self = $class->SUPER::new(
 		api_version => 3,
 		handlers => [
@@ -97,7 +103,9 @@ sub new {
 			declaration => ["html_declaration", "self,text"],
 		],
 		marked_sections => 1);
-
+  $self->{SA_character_semantics_input} = $character_semantics_input;
+  $self->{SA_encode_results} =
+    $character_semantics_input && !$character_semantics_output;
   $self;
 }
 
@@ -123,7 +131,7 @@ sub html_end {
 
   my @uri;
 
-  # add the canonified version of each uri to the detail list
+  # add the canonicalized version of each uri to the detail list
   if (defined $self->{uri}) {
     @uri = keys %{$self->{uri}};
   }
@@ -230,24 +238,42 @@ sub parse {
   # NOTE: HTML::Parser can cope with: <?xml pis>, <? with space>, so we
   # don't need to fix them here.
 
-  # HTML::Parser converts &nbsp; into a question mark ("?") for some
-  # reason, so convert them to spaces.  Confirmed in 3.31, at least.
+  # # (outdated claim) HTML::Parser converts &nbsp; into a question mark ("?")
+  # # for some reason, so convert them to spaces.  Confirmed in 3.31, at least.
+  # ... Actually it doesn't, it is correctly converted into Unicode NBSP,
+  # nevertheless it does not hurt to treat it as a space.
   $text =~ s/&nbsp;/ /g;
 
   # bug 4695: we want "<br/>" to be treated the same as "<br>", and
   # the HTML::Parser API won't do it for us
   $text =~ s/<(\w+)\s*\/>/<$1>/gi;
 
-  # Ignore stupid warning that can't be suppressed: 'Parsing of
-  # undecoded UTF-8 will give garbage when decoding entities at ..' (bug 4046)
-  {
-    local $SIG{__WARN__} = sub {
-      warn @_ unless (defined $_[0] && $_[0] =~ /^Parsing of undecoded UTF-/);
-    };
-
-    $self->SUPER::parse($text);
+  if (!$self->UNIVERSAL::can('utf8_mode')) {
+    # utf8_mode is cleared by default, only warn if it would need to be set
+    warn "message: cannot set utf8_mode, module HTML::Parser is too old\n"
+      if !$self->{SA_character_semantics_input};
+  } else {
+    $self->SUPER::utf8_mode($self->{SA_character_semantics_input} ? 0 : 1);
+    my $utf8_mode = $self->SUPER::utf8_mode;
+    dbg("message: HTML::Parser utf8_mode %s",
+        $utf8_mode ? "on (assumed UTF-8 octets)"
+                   : "off (default, assumed Unicode characters)");
   }
 
+  eval {
+    local $SIG{__WARN__} = sub {
+      my $err = $_[0];
+      $err =~ s/\s+/ /gs; $err =~ s/(.*) at .*/$1/s;
+      info("message: HTML::Parser warning: $err");
+    };
+    $self->SUPER::parse($text);
+  };
+
+  # bug 7437: deal gracefully with HTML::Parser misbehavior on unclosed <style> and <script> tags
+  # (typically from not passing the entire message to spamc, but possibly a DoS attack)
+  $self->SUPER::parse("</style>") while exists $self->{inside}{style} && $self->{inside}{style} > 0;
+  $self->SUPER::parse("</script>") while exists $self->{inside}{script} && $self->{inside}{script} > 0;
+
   $self->SUPER::eof;
 
   return $self->{text};
@@ -255,6 +281,7 @@ sub parse {
 
 sub html_tag {
   my ($self, $tag, $attr, $num) = @_;
+  utf8::encode($tag) if $self->{SA_encode_results};
 
   my $maybe_namespace = ($tag =~ m@^(?:o|st\d):[\w-]+/?$@);
 
@@ -274,15 +301,15 @@ sub html_tag {
 
   # ignore non-elements
   if (exists $elements{$tag} || exists $tricks{$tag}) {
-    text_style(@_) if exists $elements_text_style{$tag};
+    $self->text_style($tag, $attr, $num) if exists $elements_text_style{$tag};
 
     # bug 5009: things like <p> and </p> both need dealing with
-    html_whitespace(@_) if exists $elements_whitespace{$tag};
+    $self->html_whitespace($tag) if exists $elements_whitespace{$tag};
 
     # start tags
     if ($num == 1) {
-      html_uri(@_) if exists $elements_uri{$tag};
-      html_tests(@_);
+      $self->html_uri($tag, $attr) if exists $elements_uri{$tag};
+      $self->html_tests($tag, $attr, $num);
     }
     # end tags
     else {
@@ -313,13 +340,12 @@ sub push_uri {
   my ($self, $type, $uri) = @_;
 
   $uri = $self->canon_uri($uri);
+  utf8::encode($uri) if $self->{SA_encode_results};
 
   my $target = target_uri($self->{base_href} || "", $uri);
 
   # skip things like <iframe src="" ...>
-  if (length $uri) {
-    $self->{uri}->{$uri}->{types}->{$type} = 1;
-  }
+  $self->{uri}->{$uri}->{types}->{$type} = 1  if $uri ne '';
 }
 
 sub canon_uri {
@@ -351,6 +377,9 @@ sub html_uri {
     if (defined $attr->{href}) {
       $self->push_uri($tag, $attr->{href});
     }
+    if (defined $attr->{'data-saferedirecturl'}) {
+      $self->push_uri($tag, $attr->{'data-saferedirecturl'});
+    }
   }
   elsif ($tag =~ /^(?:img|frame|iframe|embed|script|bgsound)$/) {
     if (defined $attr->{src}) {
@@ -380,6 +409,7 @@ sub html_uri {
 
 	# Make sure it ends in a slash
 	$uri .= "/" unless $uri =~ m@/$@;
+        utf8::encode($uri) if $self->{SA_encode_results};
 	$self->{base_href} = $uri;
       }
     }
@@ -482,24 +512,20 @@ sub text_style {
         $new{style} = $attr->{style};
 	my @parts = split(/;/, $new{style});
 	foreach (@parts) {
-	  if (/^\s*(background-)?color:\s*(.+)\s*$/i) {
+	  if (/^\s*(background-)?color:\s*(.+?)\s*$/i) {
 	    my $whcolor = $1 ? 'bgcolor' : 'fgcolor';
 	    my $value = lc $2;
 
 	    if ($value =~ /rgb/) {
 	      $value =~ tr/0-9,//cd;
 	      my @rgb = split(/,/, $value);
-	      splice @rgb, 3;
-	      for(my $i=0; $i<3; $i++) {
-	        if (!defined $rgb[$i]) {
-	          $_ = 0;
-	        }
-	        elsif ($rgb[$i] > 255) {
-	          $rgb[$i] = 255;
-                }
-	      }
-
-              $new{$whcolor} = sprintf("#%02x%02x%02x", @rgb);
+              $new{$whcolor} = sprintf("#%02x%02x%02x",
+                                       map { !$_ ? 0 : $_ > 255 ? 255 : $_ }
+                                           @rgb[0..2]);
+            }
+            elsif ($value eq 'inherit') {
+              # do nothing, just prevent parsing of the valid
+              # CSS3 property value as 'invalid color' (Bug 7778)
             }
 	    else {
 	      $new{$whcolor} = name_to_rgb($value);
@@ -549,10 +575,10 @@ sub html_font_invisible {
 
   # invisibility
   if (substr($fg,-6) eq substr($bg,-6)) {
+    $self->put_results(font_low_contrast => 1);
     return 1;
-  }
   # near-invisibility
-  elsif ($fg =~ /^\#?([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})$/) {
+  } elsif ($fg =~ /^\#?([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})$/) {
     my ($r1, $g1, $b1) = (hex($1), hex($2), hex($3));
 
     if ($bg =~ /^\#?([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})$/) {
@@ -577,6 +603,13 @@ sub html_font_invisible {
     }
   }
 
+  
+  # invalid color
+  if ($fg eq 'invalid' or $bg eq 'invalid') {
+    $self->put_results(font_invalid_color => 1);
+    return 1;
+  }
+
   # size too small
   if ($size <= 1) {
     return 1;
@@ -598,12 +631,23 @@ sub html_tests {
   my ($self, $tag, $attr, $num) = @_;
 
   if ($tag eq "font" && exists $attr->{face}) {
-    if ($attr->{face} !~ /^[a-z][a-z -]*[a-z](?:,\s*[a-z][a-z -]*[a-z])*$/i) {
+    # Fixes from Bug 5956, 7312
+    # Examples seen in ham:
+    #  "Tahoma", Verdana, Arial, sans-serif
+    #  'Montserrat', sans-serif
+    #  Arial,Helvetica,Sans-Serif;
+    #  .SFUIDisplay
+    #  hirakakupro-w3
+    # TODO: There's still the problem completely foreign unicode strings,
+    # probably this rule should be deprecated.
+    if ($attr->{face} !~ /^\s*["'.]?[a-z ][a-z -]*[a-z]\d?["']?(?:,\s*["']?[a-z][a-z -]*[a-z]\d?["']?)*;?$/i) {
       $self->put_results(font_face_bad => 1);
     }
   }
   if ($tag eq "img" && exists $self->{inside}{a} && $self->{inside}{a} > 0) {
-    $self->{uri}->{$self->{anchor_last}}->{anchor_text}->[-1] .= "<img>\n";
+    my $uri = $self->{anchor_last};
+    utf8::encode($uri) if $self->{SA_encode_results};
+    $self->{uri}->{$uri}->{anchor_text}->[-1] .= "<img>\n";
     $self->{anchor}->[-1] .= "<img>\n";
   }
 
@@ -638,8 +682,10 @@ sub html_tests {
 
   # special text delimiters - <a> and <title>
   if ($tag eq "a") {
-    $self->{anchor_last} = (exists $attr->{href} ? $self->canon_uri($attr->{href}) : "");
-    push(@{$self->{uri}->{$self->{anchor_last}}->{anchor_text}}, '');
+    my $uri = $self->{anchor_last} =
+      (exists $attr->{href} ? $self->canon_uri($attr->{href}) : "");
+    utf8::encode($uri) if $self->{SA_encode_results};
+    push(@{$self->{uri}->{$uri}->{anchor_text}}, '');
     push(@{$self->{anchor}}, '');
   }
   if ($tag eq "title") {
@@ -680,7 +726,8 @@ sub display_text {
     }
   }
   else {
-    $text =~ s/[ \t\n\r\f\x0b\xa0]+/ /g;
+    # NBSP:  UTF-8: C2 A0, ISO-8859-*: A0
+    $text =~ s/[ \t\n\r\f\x0b]+|\xc2\xa0/ /gs;
     # trim leading whitespace if previous element was whitespace 
     # and current element is not invisible
     if (@{ $self->{text} } && !$display{invisible} &&
@@ -700,6 +747,7 @@ sub display_text {
 
 sub html_text {
   my ($self, $text) = @_;
+  utf8::encode($text) if $self->{SA_encode_results};
 
   # text that is not part of body
   if (exists $self->{inside}{script} && $self->{inside}{script} > 0)
@@ -714,7 +762,9 @@ sub html_text {
   # text that is part of body and also stored separately
   if (exists $self->{inside}{a} && $self->{inside}{a} > 0) {
     # this doesn't worry about nested anchors
-    $self->{uri}->{$self->{anchor_last}}->{anchor_text}->[-1] .= $text;
+    my $uri = $self->{anchor_last};
+    utf8::encode($uri) if $self->{SA_encode_results};
+    $self->{uri}->{$uri}->{anchor_text}->[-1] .= $text;
     $self->{anchor}->[-1] .= $text;
   }
   if (exists $self->{inside}{title} && $self->{inside}{title} > 0) {
@@ -722,7 +772,12 @@ sub html_text {
   }
 
   my $invisible_for_bayes = 0;
-  if ($text =~ /[^ \t\n\r\f\x0b\xa0]/) {
+
+  # NBSP:  UTF-8: C2 A0, ISO-8859-*: A0
+  # Bug 7374 - regex recursion limit exceeded
+  #if ($text !~ /^(?:[ \t\n\r\f\x0b]|\xc2\xa0)*\z/s) {
+  # .. alternative way, remove from string and see if there's anything left
+  if (do {(my $tmp = $text) =~ s/(?:[ \t\n\r\f\x0b]|\xc2\xa0)//gs; length($tmp)}) {
     $invisible_for_bayes = $self->html_font_invisible($text);
   }
 
@@ -730,6 +785,7 @@ sub html_text {
     # ideas discarded since they would be easy to evade:
     # 1. using \w or [A-Za-z] instead of \S or non-punctuation
     # 2. exempting certain tags
+    # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
     if ($text =~ /^[^\s\x21-\x2f\x3a-\x40\x5b-\x60\x7b-\x7e]/s &&
 	$self->{text}->[-1] =~ /[^\s\x21-\x2f\x3a-\x40\x5b-\x60\x7b-\x7e]\z/s)
     {
@@ -756,12 +812,14 @@ sub html_text {
 # note: $text includes <!-- and -->
 sub html_comment {
   my ($self, $text) = @_;
+  utf8::encode($text) if $self->{SA_encode_results};
 
   push @{ $self->{comment} }, $text;
 }
 
 sub html_declaration {
   my ($self, $text) = @_;
+  utf8::encode($text) if $self->{SA_encode_results};
 
   if ($text =~ /^<!doctype/i) {
     my $tag = "!doctype";
@@ -941,7 +999,7 @@ my %html_color = (
   yellowgreen => 0x9acd32,
 );
 
-sub name_to_rgb {
+sub name_to_rgb_old {
   my $color = lc $_[0];
 
   # note: Mozilla strips leading and trailing whitespace at this point,
@@ -971,6 +1029,89 @@ sub name_to_rgb {
   return "#" . $color;
 }
 
+sub name_to_rgb {
+  my $color = lc $_[0];
+  my $before = $color;
+
+  # strip leading and ending whitespace
+  $color =~ s/^\s*//;
+  $color =~ s/\s*$//;
+
+  # named colors
+  my $hex = $html_color{$color};
+  if (defined $hex) {
+    return sprintf("#%06x", $hex);
+  }
+
+  # IF NOT A NAME, IT SHOULD BE A HEX COLOR, HEX SHORTHAND or rgb values
+  if ($color =~ m/^[#a-f0-9]*$|rgb\([\d%, ]*\)/i) {
+
+    #Convert the RGB values to hex values so we can fall through on the programming
+
+    #RGB PERCENTS TO HEX
+    if ($color =~ m/rgb\((\d+)%,\s*(\d+)%,\s*(\d+)%\s*\)/i) {
+      $color = "#".dec2hex(int($1/100*255)).dec2hex(int($2/100*255)).dec2hex(int($3/100*255));
+    }
+
+    #RGB DEC TO HEX
+    if ($color =~ m/rgb\((\d+),\s*(\d+),\s*(\d+)\s*\)/i) {
+      $color = "#".dec2hex($1).dec2hex($2).dec2hex($3);
+    }
+
+    #PARSE THE HEX
+    if ($color =~ m/^#/) {
+      # strip to hex only
+      $color =~ s/[^a-f0-9]//ig;
+
+      # strip to 6 if greater than 6
+      if (length($color) > 6) {
+        $color=substr($color,0,6);
+      }
+
+      # strip to 3 if length < 6)
+      if (length($color) > 3 && length($color) < 6) {
+        $color=substr($color,0,3);
+      }
+
+      # pad right-hand-side to a multiple of three
+      $color .= "0" x (3 - (length($color) % 3)) if (length($color) % 3);
+
+      #DUPLICATE SHORTHAND HEX
+      if (length($color) == 3) {
+        $color =~ m/(.)(.)(.)/;
+        $color = "$1$1$2$2$3$3";
+      }
+
+    } else {
+      return "invalid";
+    } 
+
+  } else {
+    #INVALID 
+
+    #??RETURN BLACK SINCE WE DO NOT KNOW HOW THE MUA / BROWSER WILL PARSE
+    #$color = "000000";
+
+    return "invalid";
+  }
+
+  #print "DEBUG: before/after name_to_rgb new version: $before/$color\n";
+
+  return "#" . $color;
+}
+
+sub dec2hex {
+  my ($dec) = @_;
+  my ($pre) = '';
+
+  if ($dec < 16) {
+    $pre = '0';
+  }
+
+  return sprintf("$pre%lx", $dec);
+}
+
+
 use constant URI_STRICT => 0;
 
 # resolving relative URIs as defined in RFC 2396 (steps from section 5.2)
